#!/usr/bin/env python
'''
 .+"+.+"+.+"+.+"+.
(                 )
 ) rediscache.py (
(                 )
 "+.+"+.+"+.+"+.+"

rediscache is a function decorator that will cache its result in a Redis database.

Parameters:
- refresh: the number of seconds after which the cached data will be refreshed using the decorated function.
- expire: the number of seconds after which the cached data will altogether disapear from the Redis database.
- default: the value that will be returned if the data is not found in the cache. It cannot be None.
It can be bytes, string, int or float.
- enabled: This is True by default but enables to programmatically disable the cach if required.

It also reads form the environment:
- REDIS_SERVICE_HOST: the redis server host name or IP. Default is 'localhost'.
- REDIS_SERVICE_PORT: the port that the redis server listen to. Default is 6379.

Note:
A key associated to the decorated function is created using the function name and its parameters. This is based
on the value returned by the repr() function (ie: the __repr__() member) of each paramter. user defined objects
will have this function return by default a string like this:
"<amadeusbook.services.EmployeesService object at 0x7f41dedd7128>"
This will not do as each instance of the object will have a different representation no matter what. The direct
consequence will be that each key in the cache will be different, so values in cache will never be reused.
So you need to make sure that your parameters return a meaningful representation value.

TODO:
- Check what happens if redis database is getting full
'''

from functools import wraps
from json import dumps, loads
import logging
import os
import threading
from time import sleep
from types import FunctionType

import redis
from executiontime import printexecutiontime, YELLOW, RED

PREFIX = "."
REFRESH = "Refresh" # Number of times the cached function was actually called.
WAIT = "Wait" # Number of times that we executed the function in the current thread.
SLEEP = "Sleep" # Number of time that we had to wait 1s for the data to be found in the cache.
FAILED = "Failed" # Number of times the cached function raised an exception when called.
MISSED = "Missed" # Number of times the functions result was not found in the cache.
SUCCESS = "Success" # Number of times the function's result was found in the cache.
DEFAULT = "Default" # Number of times the default value was used because nothing is in the cache or the function failed.
STATS = [REFRESH, WAIT, SLEEP, FAILED, MISSED, SUCCESS, DEFAULT]

class RedisCache:
    """
    Having the decorator provided by a class allows to have some context to improve performances.
    """
    # pylint: disable=too-many-arguments

    def __init__(self, host: str = None, port: int = None, db: int = None, password: str = None, decode: bool = True, enabled: bool = True):
        self.enabled = enabled
        if self.enabled:
            # If environment variables are set for redis server, they superseed the default values.
            # But if provided at the construction, it has priority.
            if not host:
                host = os.environ.get('REDIS_SERVICE_HOST', 'localhost')
            if not port:
                port = os.environ.get('REDIS_SERVICE_PORT', 6379)
            if not db:
                db = os.environ.get('REDIS_SERVICE_DATABASE', 0)
            if not password:
                # If password is None, it is ignored.
                password = os.environ.get('REDIS_SERVICE_PASSWORD')
            self.server = redis.StrictRedis(host=host, port=port, db=db, password=password, decode_responses=decode)

    # pylint: disable=line-too-long
    def cache(self, refresh: int, expire: int, retry: int = None, default: any = '', wait: bool = False, serializer: FunctionType = None, deserializer: FunctionType = None): # NOSONAR
        '''
        Full decorator will all possible parameters. Most of the time, you should use a specialzed decorator below.

        Specific examples when to use this decorator:
        - Raw storage of byte string that you do not want to be decoded: use the decode=False.
        - JSON dumps data that doesn't need to be loaded before it is sent by a REST API: use serializer=dumps but no deserializer.
        '''

        logger = logging.getLogger(__name__)

        def decorator(function):
            '''
            The decorator itself returns a wrapper function that will replace the original one.
            '''

            @printexecutiontime('[' + function.__name__ + ']Total execution time of Redis decorator: {0}', color=YELLOW, output=logger.info)
            @wraps(function)
            def wrapper(*args, **kwargs):
                '''
                This wrapper calculates and displays the execution time of the function.
                '''

                @printexecutiontime('[' + function.__name__ + ']Execution time of call to function and storage in Redis: {0}', color=RED, output=logger.info)
                def refreshvalue(key):
                    '''
                    This gets the value provided by the function and stores it in local Redis database
                    '''
                    try:
                        # Get some stats
                        self.server.incr(REFRESH)
                        # Execute function to fetch value to cache
                        new_value = function(*args, **kwargs)
                    # pylint: disable=broad-except
                    # It is normal to catch all exception because we do not know what function is decorated.
                    except Exception as exception_in_thread:
                        # Get some stats
                        self.server.incr(FAILED)
                        # Log the error. It's not critical because maybe next time it will work.
                        logger.error("Error in Thread execution to update the Redis cache on key %s\n%s", key, exception_in_thread)
                        # Since we have no value, let's use the default
                        self.server.incr(DEFAULT)
                        new_value = default

                    # Serialize the value if requested
                    if serializer:
                        new_value = serializer(new_value)
                    # Store value in cache with expiration time
                    self.server.set(key, new_value, ex=expire)
                    # Set refresh key with refresh time
                    self.server.set(PREFIX + key, 1, ex=refresh)
                    return new_value

                def refreshvalueinthread(key):
                    '''
                    Run the refresh value in a separate thread
                    '''
                    thread = threading.Thread(target=refreshvalue, args=(key,))
                    thread.start()

                # If the cache is disabled, directly call the function
                if not self.enabled:
                    direct_value = function(*args, **kwargs)
                    # If we have decided to serialize, we always do it to be consistent
                    if serializer:
                        direct_value = serializer(direct_value)
                    if deserializer:
                        direct_value = deserializer(direct_value)
                    return direct_value

                # Lets create a key from the function's name and its parameters values
                values = ",".join([value.__str__() for value in args])
                dict_values = ",".join([str(key) + "='" + value.__str__() + "'" for key, value in kwargs.items()])
                all_args = ",".join([values, dict_values])
                key = function.__name__ + "("+ all_args + ")"

                # Get the value from the cache.
                # If it is not there we will get None.
                cached_value = self.server.get(key)

                # Time to update stats counters
                if cached_value is None:
                    self.server.incr(MISSED)
                else:
                    self.server.incr(SUCCESS)

                # If the refresh key is gone, it is time to refresh the value.
                if self.server.set(PREFIX + key, 1, ex=retry if retry else refresh, nx=True):

                    # If we found a value in the cash, we will not wait for the refresh
                    if cached_value or not wait:
                        # We just update the cache in another thread.
                        refreshvalueinthread(key)
                    else:
                        # Here we will wait, let's count it
                        self.server.incr(WAIT)
                        # We update the cash and return the value.
                        cached_value = refreshvalue(key)

                # We may still have decided to wait but another process is already getting the cache updated.
                if cached_value is None and wait:
                    # Let's wait, but this is dangerous if we never get the value in the cache.
                    # We will stop if we lose the refresh key indicating that the refreshing timed out.
                    while cached_value is None and self.server.get(PREFIX + key):
                        # Let's count how many times we wait 1s
                        self.server.incr(SLEEP)
                        sleep(1)
                        cached_value = self.server.get(key)

                # If the cache was empty, we have None in the cached_value.
                if cached_value is None:
                    # We are going to return the default value
                    self.server.incr(DEFAULT)
                    cached_value = serializer(default) if serializer else default

                # Return whatever value we have at this point.
                return deserializer(cached_value) if deserializer else cached_value

            # This allows bypassing the cache by accessing directly to the cached function
            wrapper.function = function
            return wrapper
        return decorator

    def cache_raw(self, refresh: int, expire: int, retry: int = None, default: any = ''):
        """
        Normal caching of values directly storable in redis: byte string, string, int, float.
        """
        return self.cache(refresh=refresh, expire=expire, retry=retry, default=default)

    def cache_raw_wait(self, refresh: int, expire: int, retry: int = None, default: any = ''):
        """
        Same as cache_raw() but will wait for the completion of the cached function if no value is found in redis.
        """
        return self.cache(refresh=refresh, expire=expire, retry=retry, default=default, wait=True)

    def cache_json(self, refresh: int, expire: int, retry: int = None, default: any = ''):
        """
        JSON dumps the values to be stored in redis and loads them again when returning them to the caller.
        """
        return self.cache(refresh=refresh, expire=expire, retry=retry, default=default, serializer=dumps, deserializer=loads)

    def cache_json_wait(self, refresh: int, expire: int, retry: int = None, default: any = ''):
        """
        Same as cache_json() but will wait for the completion of the cached function if no value is found in redis.
        """
        return self.cache(refresh=refresh, expire=expire, retry=retry, default=default, wait=True, serializer=dumps, deserializer=loads)

    def get_stats(self, delete=False):
        """
        Get the stats stored by RedisCache. See the list and definition at the top of this file.
        If delete is set to True we delete the stats from Redis after read.
        From Redis 6.2, it is possible to GETDEL, making sure that we do not lose some data between
        the 'get' and the 'delete'. But it is not available in the Redis (v3.5.3) python interface yet.
        """
        stats = {stat: int(self.server.get(stat) or 0) for stat in STATS}
        if delete:
            for stat in STATS:
                self.server.delete(stat)
        return stats



#!/usr/bin/env python3
"""
This will test various scenario with the redis cache.

It requires the local redis server to be started:
sudo service redis.server start

Start with unittest
python -m unittest test_rediscache.py

Start with pytest
pytest -s

Start with coverage
coverage run --source=rediscache --module pytest
coverage report --show-missing
"""
from datetime import datetime
from logging import info, INFO, basicConfig
from threading import Thread
from time import sleep
from unittest import TestCase, main

from redis import StrictRedis

from rediscache import RedisCache

# pylint: disable=missing-class-docstring, missing-function-docstring
class TestRedisCache(TestCase):

    def __init__(self, method_name: str):
        self.server = StrictRedis(decode_responses=True)
        super().__init__(method_name)

    def setUp(self):
        retour = self.server.flushdb()
        info("Redis database flushed: %s", retour)

    def test_key_in_redis(self):
        """
        This can be tested alone with:
        pytest -k test_key_in_redis
        """
        rediscache = RedisCache()
        @rediscache.cache(10, 20, wait=True)
        def func_with_args(arg, kwarg=''):
            return "%s & %s" % (arg, kwarg)
        value = func_with_args("toto", kwarg="titi")
        keys = self.server.keys("*")
        self.assertEqual(value, "toto & titi")
        self.assertIn("func_with_args(toto,kwarg='titi')", keys)

    def test_normal_cache(self):
        rediscache = RedisCache()
        @rediscache.cache_raw(1, 2)
        def my_slow_hello(name: str) -> str:
            self.server.incr('my_slow_hello')
            sleep(0.3)
            return f"Hello {name}!"

        # Ask value to go in cache
        name = 'toto'
        hello = my_slow_hello(name)
        # Make sure the Thread was started
        sleep(0.1)
        # The function was called
        self.assertEqual(self.server.get('my_slow_hello'), '1')
        # But we do not have the result yet?
        self.assertEqual(hello, '')
        # Make sure the value is in the cache
        sleep(0.5)
        # Get value from cache
        hello = my_slow_hello(name)
        self.assertEqual(hello, f"Hello {name}!")
        # Still the function has only been called once
        self.assertEqual(self.server.get('my_slow_hello'), '1')

    def test_refresh(self):
        rediscache = RedisCache()
        @rediscache.cache_raw(1, 2)
        def my_slow_hello(name: str) -> str:
            """
            A slow function to test the Redis cache.
            Each call will get a different result.
            """
            sleep(0.2)
            return f"Hello {name}! {datetime.utcnow()}"

        # Ask value to go in cache
        hello = my_slow_hello('tata')
        # Make sure we have not got it yet
        self.assertEqual(hello, '')
        # Wait for it to arrive in cache
        sleep(0.5)
        # Retrieve value from cache
        hello1 = my_slow_hello('tata')
        # Wait for expiration of value
        sleep(1)
        # Retrieve value from cache again. Should be the same. But an update is on-going.
        hello2 = my_slow_hello('tata')
        self.assertEqual(hello1, hello2)
        # Wait for cache to be updated
        sleep(0.5)
        # Retrieve value from cache again. This time it was updated.
        hello3 = my_slow_hello('tata')
        self.assertNotEqual(hello2, hello3)

    def test_default(self):
        default = 'Default'
        rediscache = RedisCache()
        @rediscache.cache_raw(1, 2, default=default)
        def my_fast_hello(name: str) -> str:
            return f"Hello {name}!"

        # First time storing value in cache
        hello = my_fast_hello('fifi')
        # Value is default the first time
        self.assertEqual(hello, default)

    def test_fail(self):
        default = 'Default'
        rediscache = RedisCache()
        @rediscache.cache_raw(1, 2, default=default)
        def my_failling_hello(name: str) -> str:
            if not name:
                raise ValueError('Invalid name')
            return f"Hello {name}!"

        # Store value in the cache
        hello = my_failling_hello('')
        # Make sure we have not got it yet
        self.assertEqual(hello, default)
        # Wait for value to be in the cache
        sleep(1.1)
        # Do we get a value from the cache?
        hello = my_failling_hello('')
        # Only the default value is in the cache because the function failed
        self.assertEqual(hello, default)

    def test_expire(self):
        default = 'Default'
        rediscache = RedisCache()
        @rediscache.cache_raw(1, 2, default=default)
        def my_fast_hello(name: str) -> str:
            return f"Hello {name}!"

        # Store value in the cache
        hello = my_fast_hello('loulou')
        # Make sure we have not got it yet
        self.assertEqual(hello, default)
        # Wait for the value to be totally expired in the cache
        sleep(3)
        # Let's try and get the value from the cache
        hello = my_fast_hello('loulou')
        # The value is not in the cache anymore
        self.assertEqual(hello, default)

    def test_empty(self):
        default = 'Default'
        rediscache = RedisCache()
        @rediscache.cache_raw(1, 2, default=default)
        def my_empty_hello(name: str) -> str:
            if name:
                return f"Hello {name}!"
            return ""

        # Store value in the cache
        hello = my_empty_hello('')
        # Make sure we have not got it yet
        self.assertEqual(hello, default)
        # Wait for the value to be stored in the cache
        sleep(0.5)
        # Let's try and get the value from the cache
        hello = my_empty_hello('')
        # We also store empty strings in the cache
        self.assertEqual(hello, '')

    def test_no_cache(self):
        rediscache = RedisCache(enabled=False)
        @rediscache.cache_raw(1, 2)
        def my_slow_hello(name: str) -> str:
            sleep(0.5)
            return f"Hello {name}!"

        # Get the value directly, no cache
        name = 'choux'
        hello = my_slow_hello(name)
        # We have the value after the first call
        self.assertEqual(hello, f"Hello {name}!")

    def test_no_cache_dumps(self):
        rediscache = RedisCache(enabled=False)
        @rediscache.cache_json(1, 2)
        def my_slow_hello(name: str) -> str:
            sleep(0.5)
            return f"Hello {name}!"

        # Get the value directly, no cache
        name = 'choux'
        hello = my_slow_hello(name)
        # We have the value after the first call
        self.assertEqual(hello, f'Hello {name}!')

    def test_very_long(self):
        rediscache = RedisCache()
        @rediscache.cache_raw(1, 10, retry=1)
        def my_very_slow_hello(name: str) -> str:
            # Count how many times the function was called
            self.server.incr('my_very_slow_hello')
            sleep(2)
            return f"Hello {name}!"

        # Stores the value in the cache
        my_very_slow_hello('hiboux')
        # Wait after the retry
        sleep(1.5)
        # Do we have it now?
        hello = my_very_slow_hello('hiboux')
        # No, we should not
        self.assertEqual(hello, '')
        # Let's see how many times the function was actually called
        sleep(0.1)
        self.assertEqual(self.server.get('my_very_slow_hello'), '2')

    def test_dict(self):
        rediscache = RedisCache()
        @rediscache.cache_json(1, 2)
        def return_dict(name: str):
            return {"hello": name}
        # Stores the value in the cache
        return_dict('you')
        # Make sure it is in the cache
        sleep(0.1)
        # Get the value from the cache
        hello = return_dict('you')
        self.assertEqual(hello, {'hello': 'you'})

    def test_dict_wait(self):
        rediscache = RedisCache()
        @rediscache.cache_json_wait(1, 2)
        def return_dict(name: str):
            sleep(0.1)
            return {"hello": name}
        # Stores the value in the cache and wait for the output
        hello = return_dict('me')
        self.assertEqual(hello, {'hello': 'me'})

    def test_wait(self):
        rediscache = RedisCache()
        @rediscache.cache_raw_wait(1, 2)
        def my_hello_wait(name: str) -> str:
            sleep(0.5)
            return f"hello {name}!"
        # Stores the value in the cache but wait for it as well
        name = 'chouchou'
        hello = my_hello_wait(name)
        self.assertEqual(hello, f"hello {name}!")

    def test_wait_thread(self):
        rediscache = RedisCache()
        @rediscache.cache_raw_wait(1, 2)
        def my_hello_wait(name: str) -> str:
            sleep(0.5)
            return f"hello {name}!"
        name = 'bob'
        # Store the value in the cache and wait in another thread
        thread = Thread(target=my_hello_wait, args=(name,))
        thread.start()
        # Make sure the thread is started but not enough that it would be completed
        sleep(0.2)
        # Now we still wait for the value
        hello = my_hello_wait(name)
        self.assertEqual(hello, f"hello {name}!")

    def test_no_decode(self):
        my_byte_string = b'This is a byte string'
        rediscache = RedisCache(decode=False)
        @rediscache.cache_raw_wait(1, 2)
        def my_bytes() -> bytes:
            sleep(0.1)
            return my_byte_string

        # Store the value in the cache and wait
        value = my_bytes()
        self.assertEqual(value, my_byte_string)
        # Wait for the value to reach the cache
        sleep(0.2)
        # Get the same value from the cache
        value = my_bytes()
        self.assertEqual(value, my_byte_string)

    def test_decorator(self):
        rediscache = RedisCache()
        @rediscache.cache(1, 2)
        def my_cached_hello(name: str) -> str:
            """This is my documentation"""
            sleep(0.1)
            return f"Hello {name}!"
        self.assertEqual(my_cached_hello.__name__, "my_cached_hello")
        self.assertEqual(my_cached_hello.__doc__, "This is my documentation")

    # This test should run first, so it needs the be the first alphabatically.
    def test_1_get_stats(self):
        rediscache = RedisCache()
        @rediscache.cache_raw_wait(1, 2)
        def function1() -> str:
            sleep(0.1)
            return "Hello function 1"
        @rediscache.cache(1, 2)
        def function2() -> str:
            sleep(0.1)
            return "Hello function 2"
        function1()
        function2()
        stats = rediscache.get_stats()
        self.assertEqual(stats["Refresh"], 2)
        self.assertEqual(stats["Wait"], 1)
        self.assertEqual(stats["Failed"], 0)
        self.assertEqual(stats["Missed"], 2)
        self.assertEqual(stats["Success"], 0)
        self.assertEqual(stats["Default"], 1)
        function1()
        function2()
        stats = rediscache.get_stats()
        self.assertEqual(stats["Refresh"], 2)
        self.assertEqual(stats["Wait"], 1)
        self.assertEqual(stats["Failed"], 0)
        self.assertEqual(stats["Missed"], 3)
        self.assertEqual(stats["Success"], 1)
        self.assertEqual(stats["Default"], 2)
        sleep(0.2)
        function1()
        function2()
        stats = rediscache.get_stats(delete=True)
        self.assertEqual(stats["Refresh"], 2)
        self.assertEqual(stats["Wait"], 1)
        self.assertEqual(stats["Failed"], 0)
        self.assertEqual(stats["Missed"], 3)
        self.assertEqual(stats["Success"], 3)
        self.assertEqual(stats["Default"], 2)
        stats = rediscache.get_stats()
        self.assertEqual(stats["Refresh"], 0)
        self.assertEqual(stats["Wait"], 0)
        self.assertEqual(stats["Failed"], 0)
        self.assertEqual(stats["Missed"], 0)
        self.assertEqual(stats["Success"], 0)
        self.assertEqual(stats["Default"], 0)

if __name__ == "__main__":
    basicConfig(level=INFO)
    main()



#!/usr/bin/env python3

"""
This allows to build the package, deploy it in PyPi and tag the source in GitHub:

> pip install -r requirements-dev.txt
> rm -rf dist
> ./setup.py sdist
> twine upload dist/*
> git tag 0.1.1
> git push origin --tags

"""

from setuptools import find_packages, setup

with open("README.md", "r") as readme:
    README = readme.read()

setup(
    name='rediscache',
    packages=find_packages(),
    version='0.1.1',
    description='Function decorator to cache results in Redis',
    long_description=README,
    long_description_content_type='text/markdown',
    # Classifiers: https://pypi.org/classifiers/
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    author='Pierre Cart-Grandjean',
    author_email='pcart-grandjean@users.noreply.github.com',
    maintainer='Amadeus IT Group',
    maintainer_email="opensource@amadeus.com",
    url='https://github.com/AmadeusITGroup/RedisCache',
    keywords=['redis', 'performance', 'cache'],
    license='MIT license',
    install_requires=[
        'redis',
        'executiontime'
    ],
)
